/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.server.mixin.server;

import net.minecraft.crash.CrashReport;
import net.minecraft.network.NetworkSystem;
import net.minecraft.network.play.server.S03PacketTimeUpdate;
import net.minecraft.profiler.Profiler;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.gui.IUpdatePlayerListBox;
import net.minecraft.server.management.ServerConfigurationManager;
import net.minecraft.util.IChatComponent;
import net.minecraft.util.ReportedException;
import net.minecraft.world.WorldServer;
import org.apache.logging.log4j.Logger;
import org.spongepowered.api.GameState;
import org.spongepowered.api.event.SpongeEventFactory;
import org.spongepowered.api.event.cause.Cause;
import org.spongepowered.api.event.game.state.GameStartedServerEvent;
import org.spongepowered.api.event.game.state.GameStoppedEvent;
import org.spongepowered.api.event.game.state.GameStoppedServerEvent;
import org.spongepowered.api.event.game.state.GameStoppingEvent;
import org.spongepowered.api.event.game.state.GameStoppingServerEvent;
import org.spongepowered.api.world.World;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Overwrite;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.asm.mixin.injection.At;
import org.spongepowered.asm.mixin.injection.Inject;
import org.spongepowered.asm.mixin.injection.callback.CallbackInfo;
import org.spongepowered.asm.mixin.injection.callback.LocalCapture;
import org.spongepowered.common.SpongeImpl;
import org.spongepowered.common.text.SpongeTexts;
import org.spongepowered.server.SpongeVanilla;
import org.spongepowered.server.world.VanillaDimensionManager;

import java.util.Hashtable;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.FutureTask;

@Mixin(MinecraftServer.class)
public abstract class MixinMinecraftServer {

    @Shadow private static Logger logger;
    @Shadow protected Queue<?> futureTaskQueue;
    @Shadow private ServerConfigurationManager serverConfigManager;
    @Shadow private Profiler theProfiler;
    @Shadow private int tickCounter;
    @Shadow abstract boolean getAllowNether();
    @Shadow abstract NetworkSystem getNetworkSystem();
    @Shadow List<?> playersOnline;

    private Hashtable<Integer, long[]> worldTickTimes = new Hashtable<Integer, long[]>();

    @Inject(method = "run", at = @At(value = "INVOKE", target = "Lnet/minecraft/server/MinecraftServer;addFaviconToStatusResponse"
            + "(Lnet/minecraft/network/ServerStatusResponse;)V", shift = At.Shift.AFTER))
    public void callServerStarted(CallbackInfo ci) {
        SpongeVanilla.INSTANCE.postState(GameStartedServerEvent.class, GameState.SERVER_STARTED);
    }

    @Inject(method = "run", at = @At(value = "INVOKE", target = "Lnet/minecraft/server/MinecraftServer;finalTick(Lnet/minecraft/crash/CrashReport;)V",
            ordinal = 0, shift = At.Shift.BY, by = -9))
    public void callServerStopping(CallbackInfo ci) {
        SpongeVanilla.INSTANCE.postState(GameStoppingServerEvent.class, GameState.SERVER_STOPPING);
    }

    @Inject(method = "run", at = @At(value = "INVOKE", target = "Lnet/minecraft/server/MinecraftServer;systemExitNow()V"))
    public void callServerStopped(CallbackInfo ci) {
        SpongeVanilla.INSTANCE.postState(GameStoppedServerEvent.class, GameState.SERVER_STOPPED);
        SpongeVanilla.INSTANCE.postState(GameStoppingEvent.class, GameState.GAME_STOPPING);
        SpongeVanilla.INSTANCE.postState(GameStoppedEvent.class, GameState.GAME_STOPPED);
    }

    @Inject(method = "stopServer", at = @At(value = "INVOKE", target = "Lnet/minecraft/world/WorldServer;flush()V"),
            locals = LocalCapture.CAPTURE_FAILHARD)
    public void callWorldUnload(CallbackInfo ci, int i, WorldServer worldserver) {
        SpongeImpl.getGame().getEventManager().post(SpongeEventFactory.createUnloadWorldEvent(SpongeImpl.getGame(), Cause.of(this), (World) worldserver));
    }

    @Overwrite
    public void updateTimeLightAndEntities() {
        this.theProfiler.func_76320_a("jobs");
        Queue<?> queue = this.futureTaskQueue;

        synchronized (this.futureTaskQueue) {
            while (!this.futureTaskQueue.isEmpty()) {
                try {
                    ((FutureTask) this.futureTaskQueue.poll()).run();
                } catch (Throwable throwable2) {
                    logger.fatal(throwable2);
                }
            }
        }

        this.theProfiler.func_76318_c("levels");
        int j;

        Integer[] ids = VanillaDimensionManager.getIDs(this.tickCounter % 200 == 0);

        for (j = 0; j < ids.length; ++j) {
            int id = ids[j];
            long i = System.nanoTime();

            if (j == 0 || this.getAllowNether()) {
                WorldServer worldserver = VanillaDimensionManager.getWorldFromDimId(id);
                this.theProfiler.func_76320_a(worldserver.func_72912_H().func_76065_j());

                if (this.tickCounter % 20 == 0) {
                    this.theProfiler.func_76320_a("timeSync");
                    this.serverConfigManager.func_148537_a(
                            new S03PacketTimeUpdate(worldserver.func_82737_E(), worldserver.func_72820_D(),
                                    worldserver.func_82736_K().func_82766_b("doDaylightCycle")), worldserver.field_73011_w.func_177502_q());
                    this.theProfiler.func_76319_b();
                }

                this.theProfiler.func_76320_a("tick");
                CrashReport crashreport;

                try {
                    worldserver.func_72835_b();
                } catch (Throwable throwable1) {
                    crashreport = CrashReport.func_85055_a(throwable1, "Exception ticking world");
                    worldserver.func_72914_a(crashreport);
                    throw new ReportedException(crashreport);
                }

                try {
                    worldserver.func_72939_s();
                } catch (Throwable throwable) {
                    crashreport = CrashReport.func_85055_a(throwable, "Exception ticking world entities");
                    worldserver.func_72914_a(crashreport);
                    throw new ReportedException(crashreport);
                }

                this.theProfiler.func_76319_b();
                this.theProfiler.func_76320_a("tracker");
                worldserver.func_73039_n().func_72788_a();
                this.theProfiler.func_76319_b();
                this.theProfiler.func_76319_b();
            }
            this.worldTickTimes.get(id)[this.tickCounter % 100] = System.nanoTime() - i;
        }
        this.theProfiler.func_76318_c("dim_unloading");
        VanillaDimensionManager.unloadWorlds(this.worldTickTimes);
        this.theProfiler.func_76318_c("connection");
        this.getNetworkSystem().func_151269_c();
        this.theProfiler.func_76318_c("players");
        this.serverConfigManager.func_72374_b();
        this.theProfiler.func_76318_c("tickables");

        for (j = 0; j < this.playersOnline.size(); ++j) {
            ((IUpdatePlayerListBox) this.playersOnline.get(j)).func_73660_a();
        }

        this.theProfiler.func_76319_b();
    }

    @Overwrite
    public String getServerModName() {
        return SpongeVanilla.INSTANCE.getName();
    }

    @Overwrite
    public void addChatMessage(IChatComponent component) {
        logger.info(SpongeTexts.toLegacy(component));
    }

    public Hashtable<Integer, long[]> getWorldTickTimes() {
        return this.worldTickTimes;
    }

}
